@@ -9,7 +9,9 @@
 
 import asyncio
 import copy
+import functools
 import io
+import logging
 import math
 import os
 import sys
@@ -21,6 +23,7 @@ from typing import Any, Callable, Generic, List, Optional, Tuple, TypeVar, Union
 
 import torch
 from torch.distributed._shard.sharded_tensor import (
+    Shard as ShardedTensorShard,
     ShardedTensor,
     ShardedTensorMetadata,
     ShardMetadata,
@@ -55,6 +58,8 @@ from .torch_dist_checkpoint.metadata import (
 )
 from .torch_dist_checkpoint.resharding import prepare_sharded_tensor_read
 
+logger: logging.Logger = logging.getLogger(__name__)
+
 
 @dataclass
 class Chunk:
@@ -108,7 +113,13 @@ class ChunkedTensorIOPreparer:
 
     @classmethod
     def prepare_write(
-        cls, storage_path: str, tensor: torch.Tensor, chunking_instruction: List[Chunk]
+        cls,
+        storage_path: str,
+        tensor: torch.Tensor,
+        chunking_instruction: List[Chunk],
+        _tensor_prepare_func: Optional[
+            Callable[[torch.Tensor, bool], torch.Tensor]
+        ] = None,
     ) -> Tuple[ChunkedTensorEntry, List[WriteReq]]:
         write_reqs = []
         chunks = []
@@ -191,6 +202,9 @@ class ShardedTensorIOPreparer:
         cls,
         storage_path: str,
         obj: ShardedTensor,
+        _tensor_prepare_func: Optional[
+            Callable[[torch.Tensor, bool], torch.Tensor]
+        ] = None,
     ) -> Tuple[ShardedTensorEntry, List[WriteReq]]:
         shards = []
         write_reqs = []
@@ -212,7 +226,9 @@ class ShardedTensorIOPreparer:
             for tensor, offsets, sizes in subdivided:
                 suffix = "_".join(str(i) for i in offsets)
                 entry, tensor_write_reqs = TensorIOPreparer.prepare_write(
-                    storage_path=f"{storage_path}_{suffix}", tensor=tensor
+                    storage_path=f"{storage_path}_{suffix}",
+                    tensor=tensor,
+                    _tensor_prepare_func=_tensor_prepare_func,
                 )
                 write_reqs += tensor_write_reqs
 
@@ -225,11 +241,34 @@ class ShardedTensorIOPreparer:
                 )
         return ShardedTensorEntry(shards=shards), write_reqs
 
+    @staticmethod
+    def _get_global_shape(entry: ShardedTensorEntry) -> List[int]:
+        # TODO: the global dtype and shape should be tracked in the metadata
+        global_shape = [0] * len(entry.shards[0].sizes)
+        for shard in entry.shards:
+            for dim in range(len(shard.offsets)):
+                if shard.offsets[dim] + shard.sizes[dim] > global_shape[dim]:
+                    global_shape[dim] = shard.offsets[dim] + shard.sizes[dim]
+        return global_shape
+
+    @staticmethod
+    def _validate_shape(global_shape: List[int], obj_out: torch.Tensor) -> None:
+        if isinstance(obj_out, ShardedTensor):
+            out_shape = list(obj_out.metadata().size)
+        else:
+            out_shape = list(obj_out.shape)
+        if out_shape != global_shape:
+            logger.warn(
+                "The shape of obj_out ({out_shape}) is different from the "
+                "shape of the persisted sharded tensor ({global_shape}). "
+                "Only the overlapping part will be loaded. "
+            )
+
     @classmethod
     def prepare_read(
         cls,
         entry: ShardedTensorEntry,
-        obj_out: Optional[ShardedTensor] = None,
+        obj_out: Optional[torch.Tensor] = None,
     ) -> List[ReadReq]:
         if obj_out is None:
             # TODO: support loading sharded tensor without obj_out
@@ -237,6 +276,22 @@ class ShardedTensorIOPreparer:
                 "Reading a ShardedTensor without a runtime object is not yet supported."
             )
 
+        global_shape = cls._get_global_shape(entry=entry)
+        cls._validate_shape(global_shape=global_shape, obj_out=obj_out)
+        if isinstance(obj_out, ShardedTensor):
+            local_shards_out = obj_out.local_shards()
+        else:
+            local_shards_out = [
+                ShardedTensorShard(
+                    tensor=obj_out,
+                    metadata=ShardMetadata(
+                        shard_offsets=[0] * len(obj_out.shape),
+                        shard_sizes=list(obj_out.shape),
+                        placement="cpu",  # Unused for load
+                    ),
+                )
+            ]
+
         metadata = ExtendedTensorMetadata(
             tensor_metadata=ShardedTensorMetadata(),  # Unused for load
             storage_metadata=[],
@@ -255,7 +310,7 @@ class ShardedTensorIOPreparer:
                 )
             )
         tensor_read_reqs = prepare_sharded_tensor_read(
-            metadata=metadata, sharded_tensor_out=obj_out
+            metadata=metadata, local_shards_out=local_shards_out
         )
         locations_to_load = {twr.storage_key for twr in tensor_read_reqs}
 
@@ -331,11 +386,20 @@ def tensor_to_cpu(tensor: torch.Tensor) -> torch.Tensor:
 
 
 class TensorBufferStager(BufferStager):
-    def __init__(self, tensor: torch.Tensor, entry: TensorEntry) -> None:
+    def __init__(
+        self,
+        tensor: torch.Tensor,
+        entry: TensorEntry,
+        _tensor_prepare_func: Callable[[torch.Tensor, bool], torch.Tensor],
+    ) -> None:
         self.tensor = tensor
         self.entry = entry
+        self._tensor_prepare_func = _tensor_prepare_func
 
     async def stage_buffer(self, executor: Optional[Executor] = None) -> BufferType:
+        # TODO: if the custom prepared tensor is different from the original
+        # tensor and is a CPU tensor, don't copy it.
+        self.tensor = self._tensor_prepare_func(self.tensor, False)  # tracing=False
         if self.tensor.is_cuda:
             # It would be nice to copy from GPU via DMA. However, it is very
             # difficult to figure out the safe amount of page-locked memory
@@ -417,12 +481,32 @@ class TensorBufferConsumer(BufferConsumer):
             raise ValueError(f"Unrecognized serializer: {self.entry.serializer}.")
 
 
+def _identity_tensor_prepare_func(
+    path: str, tensor: torch.Tensor, tracing: bool
+) -> torch.Tensor:
+    return tensor
+
+
 class TensorIOPreparer:
     @staticmethod
     def prepare_write(
-        storage_path: str, tensor: torch.Tensor
+        storage_path: str,
+        tensor: torch.Tensor,
+        _tensor_prepare_func: Optional[
+            Callable[[torch.Tensor, bool], torch.Tensor]
+        ] = None,
     ) -> Tuple[TensorEntry, List[WriteReq]]:
-        if tensor.dtype in BUFFER_PROTOCOL_SUPPORTED_DTYPES:
+        if not _tensor_prepare_func:
+            _tensor_prepare_func = functools.partial(_identity_tensor_prepare_func, "")
+
+        proc_tensor = _tensor_prepare_func(tensor, True)  # tracing=True
+        if proc_tensor.shape != tensor.shape:
+            raise RuntimeError(
+                "_tensor_prepare_func shouldn't change the tensor's shape "
+                f"(changed from {tensor.shape} to {proc_tensor.shape})."
+            )
+
+        if proc_tensor.dtype in BUFFER_PROTOCOL_SUPPORTED_DTYPES:
             serializer = Serializer.BUFFER_PROTOCOL.value
         else:
             serializer = Serializer.TORCH_SAVE.value
@@ -430,11 +514,16 @@ class TensorIOPreparer:
         entry = TensorEntry(
             location=storage_path,
             serializer=serializer,
-            dtype=dtype_to_string(tensor.dtype),
-            shape=list(tensor.shape),
+            dtype=dtype_to_string(proc_tensor.dtype),
+            shape=list(proc_tensor.shape),
             replicated=False,
         )
-        buffer_stager = TensorBufferStager(tensor=tensor, entry=entry)
+        # stage the actual tensor, not processed tensor
+        buffer_stager = TensorBufferStager(
+            tensor=tensor,
+            entry=entry,
+            _tensor_prepare_func=_tensor_prepare_func,
+        )
         return entry, [WriteReq(path=storage_path, buffer_stager=buffer_stager)]
 
     @classmethod
@@ -444,6 +533,11 @@ class TensorIOPreparer:
         tensor_out: Optional[torch.Tensor] = None,
         buffer_size_limit_bytes: Optional[int] = None,
     ) -> List[ReadReq]:
+        # TODO: When the output tensor is a CPU tensor, we should directly load
+        # into its storage buffer. This is an important optimization because:
+        # - We eliminate an allocation and a copy
+        # - With the extra allocation, the I/O concurrency will be severely
+        # reduced by the scheduler in a memory constrained environment
         if tensor_out is None:
             raise RuntimeError(
                 "Reading a Tensor without a runtime object is not yet supported."
@@ -582,6 +676,7 @@ def prepare_write(
     logical_path: str,
     rank: int,
     replicated: bool,
+    _tensor_prepare_func: Optional[Callable[[torch.Tensor, bool], torch.Tensor]] = None,
 ) -> Tuple[Entry, List[WriteReq]]:
     """
     Prepare write for an object.
@@ -598,9 +693,13 @@ def prepare_write(
     """
     storage_path = get_storage_path(obj, logical_path, rank, replicated)
     if isinstance(obj, ShardedTensor):
-        return ShardedTensorIOPreparer.prepare_write(storage_path, obj)
+        return ShardedTensorIOPreparer.prepare_write(
+            storage_path, obj, _tensor_prepare_func
+        )
     elif isinstance(obj, torch.Tensor):
-        entry, obj_write_req = TensorIOPreparer.prepare_write(storage_path, obj)
+        entry, obj_write_req = TensorIOPreparer.prepare_write(
+            storage_path, obj, _tensor_prepare_func
+        )
     else:
         entry, obj_write_req = ObjectIOPreparer.prepare_write(storage_path, obj)
 
