@@ -21,14 +21,13 @@ from datetime import timedelta
 from functools import reduce
 from operator import mul
 from threading import Thread
-from typing import Any, cast, Dict, List, Optional, Tuple, TypeVar
+from typing import Any, Callable, cast, Dict, List, Optional, Tuple, TypeVar
 
 import numpy as np
 import torch
 import torch.distributed as dist
 from torch.distributed._shard.sharded_tensor import ShardedTensor
 from torch.nn.parallel import DistributedDataParallel as DDP
-from torchsnapshot.io_preparer import Chunk
 from torchsnapshot.manifest import ChunkedTensorEntry
 from torchsnapshot.serialization import dtype_to_element_size, string_to_dtype
 
@@ -36,6 +35,8 @@ from .dist_store import get_or_create_store, LinearBarrier
 
 from .flatten import flatten, inflate
 from .io_preparer import (
+    _identity_tensor_prepare_func,
+    Chunk,
     ChunkedTensorIOPreparer,
     get_storage_path,
     ObjectBufferConsumer,
@@ -176,6 +177,9 @@ class Snapshot:
         app_state: AppState,
         pg: Optional[dist.ProcessGroup] = None,
         replicated: Optional[List[str]] = None,
+        _custom_tensor_prepare_func: Optional[
+            Callable[[str, torch.Tensor, bool], torch.Tensor]
+        ] = None,
     ) -> "Snapshot":
         """
         Take a snapshot from the program state.
@@ -195,8 +199,11 @@ class Snapshot:
             The newly taken snapshot.
         """
         torch._C._log_api_usage_once("torchsnapshot.Snapshot.take")
+        cls._validate_app_state(app_state)
+
         event_loop = asyncio.new_event_loop()
         pg_wrapper = PGWrapper(pg=pg)
+
         path, replicated = cls._coalesce_path_and_replicated(
             path=path,
             pg_wrapper=pg_wrapper,
@@ -213,6 +220,7 @@ class Snapshot:
             pg_wrapper=PGWrapper(pg),
             storage=storage,
             event_loop=event_loop,
+            _custom_tensor_prepare_func=_custom_tensor_prepare_func,
         )
         pending_io_work.sync_complete(event_loop=event_loop)
 
@@ -238,6 +246,9 @@ class Snapshot:
         app_state: AppState,
         pg: Optional[dist.ProcessGroup] = None,
         replicated: Optional[List[str]] = None,
+        _custom_tensor_prepare_func: Optional[
+            Callable[[str, torch.Tensor, bool], torch.Tensor]
+        ] = None,
     ) -> "PendingSnapshot":
         """
         Asynchronously take a snapshot from the program state.
@@ -265,6 +276,8 @@ class Snapshot:
             invoked.
         """
         torch._C._log_api_usage_once("torchsnapshot.Snapshot.async_take")
+        cls._validate_app_state(app_state)
+
         event_loop = asyncio.new_event_loop()
         pg_wrapper = PGWrapper(pg=pg)
         path, replicated = cls._coalesce_path_and_replicated(
@@ -276,6 +289,7 @@ class Snapshot:
         storage = url_to_storage_plugin_in_event_loop(
             url_path=path, event_loop=event_loop
         )
+
         pending_io_work, metadata = cls._take_impl(
             path=path,
             app_state=app_state,
@@ -283,6 +297,7 @@ class Snapshot:
             pg_wrapper=PGWrapper(pg),
             storage=storage,
             event_loop=event_loop,
+            _custom_tensor_prepare_func=_custom_tensor_prepare_func,
         )
         # PendingSnapshot is responsible for closing `storage` and `event_loop`
         return PendingSnapshot(
@@ -303,13 +318,17 @@ class Snapshot:
         pg_wrapper: PGWrapper,
         storage: StoragePlugin,
         event_loop: asyncio.AbstractEventLoop,
+        _custom_tensor_prepare_func: Optional[
+            Callable[[str, torch.Tensor, bool], torch.Tensor]
+        ] = None,
     ) -> Tuple[PendingIOWork, SnapshotMetadata]:
-        # TODO: validate app_state
-
         app_state = app_state.copy()
         rng_state_item = cls._pop_rng_state(app_state=app_state)
         rng_state_dict = None
 
+        if not _custom_tensor_prepare_func:
+            _custom_tensor_prepare_func = _identity_tensor_prepare_func
+
         manifest: Manifest = {}
         flattened: Dict[str, Any] = {}
 
@@ -348,9 +367,7 @@ class Snapshot:
         # function won't affect the RNG state or execute application code.
         if rng_state_item is not None:
             _, stateful = rng_state_item
-            rng_state_dict = stateful.load_state_dict(
-                cast(Dict[str, Any], rng_state_dict)  # pyre-ignore[33]
-            )
+            stateful.load_state_dict(cast(Dict[str, torch.Tensor], rng_state_dict))
 
         replicated_paths = cls._calculate_replicated_entries(
             flattened, replicated, pg_wrapper
@@ -381,6 +398,9 @@ class Snapshot:
                 ),
                 tensor=flattened[logical_path],
                 chunking_instruction=chunking_instructions[logical_path],
+                _tensor_prepare_func=functools.partial(
+                    _custom_tensor_prepare_func, logical_path
+                ),
             )
             entry.replicated = logical_path in replicated_set
             object_entries[logical_path] = entry
@@ -392,6 +412,9 @@ class Snapshot:
                 logical_path=logical_path,
                 rank=pg_wrapper.get_rank(),
                 replicated=logical_path in replicated_set,
+                _tensor_prepare_func=functools.partial(
+                    _custom_tensor_prepare_func, logical_path
+                ),
             )
             object_entries[logical_path] = entry
             write_reqs.extend(item_write_reqs)
@@ -422,6 +445,8 @@ class Snapshot:
             app_state: The program state to restore from the snapshot.
         """
         torch._C._log_api_usage_once("torchsnapshot.Snapshot.restore")
+        self._validate_app_state(app_state)
+
         event_loop = asyncio.new_event_loop()
         pg_wrapper = PGWrapper(self.pg)
         rank = pg_wrapper.get_rank()
@@ -500,8 +525,9 @@ class Snapshot:
         "sharded".
 
         If the persisted object is a sharded tensor, `obj_out` must be
-        supplied. `read_object` will correctly populate `obj_out`'s local
-        shards according to its sharding spec.
+        supplied. The supplied tensor can be either a tensor or sharded tensor.
+        `read_object` will correctly populate `obj_out`'s data according to
+        sharding spec.
 
         Args:
             path: The path to the persisted object.
@@ -622,6 +648,16 @@ class Snapshot:
         replicated_paths = replicated_paths_list[0]
         return replicated_paths
 
+    @staticmethod
+    def _validate_app_state(app_state: AppState) -> None:
+        # performs runtime typechecking that all values are Stateful
+        for key, value in app_state.items():
+            if not isinstance(value, Stateful):
+                value_type = type(value)
+                raise TypeError(
+                    f"Expected Stateful in app_state for key {key}, got {value_type}."
+                )
+
     @classmethod
     def _load_stateful(
         cls,
